"""This module will contain personal modules from our coworkers, allowing for easier sharing
of data processing files that utilize functions that might not otherwise be of general use.
"""

from . import marco, nick, vlad

__all__ = ["nick", "vlad", "marco"]
