"""Contains functions for plotting, and for formatting plots, with the goal of more easily obtaining a consistent style across plots made by different co-authors."""

from .plot import Char, label_letter, set_style, showmo

__all__ = ["showmo", "set_style", "label_letter", "Char"]
